@@ -17,6 +17,22 @@ description:
   - For Windows targets, use the M(ansible.windows.win_uri) module instead.
 version_added: "1.1"
 options:
+  ciphers:
+    description:
+      - SSL/TLS Ciphers to use for the request.
+      - 'When a list is provided, all ciphers are joined in order with C(:)'
+      - See the L(OpenSSL Cipher List Format,https://www.openssl.org/docs/manmaster/man1/openssl-ciphers.html#CIPHER-LIST-FORMAT)
+        for more details.
+      - The available ciphers is dependent on the Python and OpenSSL/LibreSSL versions
+    type: list
+    elements: str
+    version_added: '2.14'
+  decompress:
+    description:
+      - Whether to attempt to decompress gzip content-encoded responses
+    type: bool
+    default: true
+    version_added: '2.14'
   url:
     description:
       - HTTP or HTTPS URL in the form (http|https)://host.domain[:port]/path
@@ -88,8 +104,8 @@ options:
       - Whether or not the URI module should follow redirects. C(all) will follow all redirects.
         C(safe) will follow only "safe" redirects, where "safe" means that the client is only
         doing a GET or HEAD on the URI to which it is being redirected. C(none) will not follow
-        any redirects. Note that C(yes) and C(no) choices are accepted for backwards compatibility,
-        where C(yes) is the equivalent of C(all) and C(no) is the equivalent of C(safe). C(yes) and C(no)
+        any redirects. Note that C(true) and C(false) choices are accepted for backwards compatibility,
+        where C(true) is the equivalent of C(all) and C(false) is the equivalent of C(safe). C(true) and C(false)
         are deprecated and will be removed in some future version of Ansible.
     type: str
     choices: ['all', 'no', 'none', 'safe', 'urllib2', 'yes']
@@ -122,9 +138,9 @@ options:
     version_added: '2.1'
   validate_certs:
     description:
-      - If C(no), SSL certificates will not be validated.
-      - This should only set to C(no) used on personally controlled sites using self-signed certificates.
-      - Prior to 1.9.2 the code defaulted to C(no).
+      - If C(false), SSL certificates will not be validated.
+      - This should only set to C(false) used on personally controlled sites using self-signed certificates.
+      - Prior to 1.9.2 the code defaulted to C(false).
     type: bool
     default: yes
     version_added: '1.9.2'
@@ -149,23 +165,24 @@ options:
     description:
       - Path to file to be submitted to the remote server.
       - Cannot be used with I(body).
+      - Should be used with I(force_basic_auth) to ensure success when the remote end sends a 401.
     type: path
     version_added: '2.7'
   remote_src:
     description:
-      - If C(no), the module will search for the C(src) on the controller node.
-      - If C(yes), the module will search for the C(src) on the managed (remote) node.
+      - If C(false), the module will search for the C(src) on the controller node.
+      - If C(true), the module will search for the C(src) on the managed (remote) node.
     type: bool
     default: no
     version_added: '2.7'
   force:
     description:
-      - If C(yes) do not get a cached copy.
+      - If C(true) do not get a cached copy.
     type: bool
     default: no
   use_proxy:
     description:
-      - If C(no), it will not use a proxy, even if one is defined in an environment variable on the target hosts.
+      - If C(false), it will not use a proxy, even if one is defined in an environment variable on the target hosts.
     type: bool
     default: yes
   unix_socket:
@@ -198,6 +215,14 @@ options:
     type: bool
     default: no
     version_added: '2.11'
+  use_netrc:
+    description:
+      - Determining whether to use credentials from ``~/.netrc`` file
+      - By default .netrc is used with Basic authentication headers
+      - When set to False, .netrc credentials are ignored
+    type: bool
+    default: true
+    version_added: '2.14'
 extends_documentation_fragment:
   - action_common_attributes
   - files
@@ -335,44 +360,25 @@ EXAMPLES = r'''
   retries: 720 # 720 * 5 seconds = 1hour (60*60/5)
   delay: 5 # Every 5 seconds
 
-# There are issues in a supporting Python library that is discussed in
-# https://github.com/ansible/ansible/issues/52705 where a proxy is defined
-# but you want to bypass proxy use on CIDR masks by using no_proxy
-- name: Work around a python issue that doesn't support no_proxy envvar
-  ansible.builtin.uri:
-    follow_redirects: none
-    validate_certs: false
-    timeout: 5
-    url: "http://{{ ip_address }}:{{ port | default(80) }}"
-  register: uri_data
-  failed_when: false
-  changed_when: false
-  vars:
-    ip_address: 192.0.2.1
-  environment: |
-      {
-        {% for no_proxy in (lookup('ansible.builtin.env', 'no_proxy') | regex_replace('\s*,\s*', ' ') ).split() %}
-          {% if no_proxy | regex_search('\/') and
-                no_proxy | ipaddr('net') != '' and
-                no_proxy | ipaddr('net') != false and
-                ip_address | ipaddr(no_proxy) is not none and
-                ip_address | ipaddr(no_proxy) != false %}
-            'no_proxy': '{{ ip_address }}'
-          {% elif no_proxy | regex_search(':') != '' and
-                  no_proxy | regex_search(':') != false and
-                  no_proxy == ip_address + ':' + (port | default(80)) %}
-            'no_proxy': '{{ ip_address }}:{{ port | default(80) }}'
-          {% elif no_proxy | ipaddr('host') != '' and
-                  no_proxy | ipaddr('host') != false and
-                  no_proxy == ip_address %}
-            'no_proxy': '{{ ip_address }}'
-          {% elif no_proxy | regex_search('^(\*|)\.') != '' and
-                  no_proxy | regex_search('^(\*|)\.') != false and
-                  no_proxy | regex_replace('\*', '') in ip_address %}
-            'no_proxy': '{{ ip_address }}'
-          {% endif %}
-        {% endfor %}
-      }
+- name: Provide SSL/TLS ciphers as a list
+  uri:
+    url: https://example.org
+    ciphers:
+      - '@SECLEVEL=2'
+      - ECDH+AESGCM
+      - ECDH+CHACHA20
+      - ECDH+AES
+      - DHE+AES
+      - '!aNULL'
+      - '!eNULL'
+      - '!aDSS'
+      - '!SHA1'
+      - '!AESCCM'
+
+- name: Provide SSL/TLS ciphers as an OpenSSL formatted cipher list
+  uri:
+    url: https://example.org
+    ciphers: '@SECLEVEL=2:ECDH+AESGCM:ECDH+CHACHA20:ECDH+AES:DHE+AES:!aNULL:!eNULL:!aDSS:!SHA1:!AESCCM'
 '''
 
 RETURN = r'''
@@ -457,61 +463,39 @@ def format_message(err, resp):
 
 
 def write_file(module, dest, content, resp):
-    # create a tempfile with some test content
-    fd, tmpsrc = tempfile.mkstemp(dir=module.tmpdir)
-    f = os.fdopen(fd, 'wb')
+    """
+    Create temp file and write content to dest file only if content changed
+    """
+
+    tmpsrc = None
+
     try:
-        if isinstance(content, binary_type):
-            f.write(content)
-        else:
-            shutil.copyfileobj(content, f)
+        fd, tmpsrc = tempfile.mkstemp(dir=module.tmpdir)
+        with os.fdopen(fd, 'wb') as f:
+            if isinstance(content, binary_type):
+                f.write(content)
+            else:
+                shutil.copyfileobj(content, f)
     except Exception as e:
-        os.remove(tmpsrc)
+        if tmpsrc and os.path.exists(tmpsrc):
+            os.remove(tmpsrc)
         msg = format_message("Failed to create temporary content file: %s" % to_native(e), resp)
         module.fail_json(msg=msg, **resp)
-    f.close()
 
-    checksum_src = None
-    checksum_dest = None
-
-    # raise an error if there is no tmpsrc file
-    if not os.path.exists(tmpsrc):
-        os.remove(tmpsrc)
-        msg = format_message("Source '%s' does not exist" % tmpsrc, resp)
-        module.fail_json(msg=msg, **resp)
-    if not os.access(tmpsrc, os.R_OK):
-        os.remove(tmpsrc)
-        msg = format_message("Source '%s' not readable" % tmpsrc, resp)
-        module.fail_json(msg=msg, **resp)
     checksum_src = module.sha1(tmpsrc)
-
-    # check if there is no dest file
-    if os.path.exists(dest):
-        # raise an error if copy has no permission on dest
-        if not os.access(dest, os.W_OK):
-            os.remove(tmpsrc)
-            msg = format_message("Destination '%s' not writable" % dest, resp)
-            module.fail_json(msg=msg, **resp)
-        if not os.access(dest, os.R_OK):
-            os.remove(tmpsrc)
-            msg = format_message("Destination '%s' not readable" % dest, resp)
-            module.fail_json(msg=msg, **resp)
-        checksum_dest = module.sha1(dest)
-    else:
-        if not os.access(os.path.dirname(dest), os.W_OK):
-            os.remove(tmpsrc)
-            msg = format_message("Destination dir '%s' not writable" % os.path.dirname(dest), resp)
-            module.fail_json(msg=msg, **resp)
+    checksum_dest = module.sha1(dest)
 
     if checksum_src != checksum_dest:
         try:
-            shutil.copyfile(tmpsrc, dest)
+            module.atomic_move(tmpsrc, dest)
         except Exception as e:
-            os.remove(tmpsrc)
+            if os.path.exists(tmpsrc):
+                os.remove(tmpsrc)
             msg = format_message("failed to copy %s to %s: %s" % (tmpsrc, dest, to_native(e)), resp)
             module.fail_json(msg=msg, **resp)
 
-    os.remove(tmpsrc)
+    if os.path.exists(tmpsrc):
+        os.remove(tmpsrc)
 
 
 def absolute_location(url, location):
@@ -568,7 +552,8 @@ def form_urlencoded(body):
     return body
 
 
-def uri(module, url, dest, body, body_format, method, headers, socket_timeout, ca_path, unredirected_headers):
+def uri(module, url, dest, body, body_format, method, headers, socket_timeout, ca_path, unredirected_headers, decompress,
+        ciphers, use_netrc):
     # is dest is set and is a directory, let's check if we get redirected and
     # set the filename from that url
 
@@ -592,7 +577,8 @@ def uri(module, url, dest, body, body_format, method, headers, socket_timeout, c
     resp, info = fetch_url(module, url, data=data, headers=headers,
                            method=method, timeout=socket_timeout, unix_socket=module.params['unix_socket'],
                            ca_path=ca_path, unredirected_headers=unredirected_headers,
-                           **kwargs)
+                           use_proxy=module.params['use_proxy'], decompress=decompress,
+                           ciphers=ciphers, use_netrc=use_netrc, **kwargs)
 
     if src:
         # Try to close the open file handle
@@ -625,6 +611,9 @@ def main():
         remote_src=dict(type='bool', default=False),
         ca_path=dict(type='path', default=None),
         unredirected_headers=dict(type='list', elements='str', default=[]),
+        decompress=dict(type='bool', default=True),
+        ciphers=dict(type='list', elements='str'),
+        use_netrc=dict(type='bool', default=True),
     )
 
     module = AnsibleModule(
@@ -646,6 +635,9 @@ def main():
     ca_path = module.params['ca_path']
     dict_headers = module.params['headers']
     unredirected_headers = module.params['unredirected_headers']
+    decompress = module.params['decompress']
+    ciphers = module.params['ciphers']
+    use_netrc = module.params['use_netrc']
 
     if not re.match('^[A-Z]+$', method):
         module.fail_json(msg="Parameter 'method' needs to be a single word in uppercase, like GET or POST.")
@@ -688,7 +680,8 @@ def main():
     # Make the request
     start = datetime.datetime.utcnow()
     r, info = uri(module, url, dest, body, body_format, method,
-                  dict_headers, socket_timeout, ca_path, unredirected_headers)
+                  dict_headers, socket_timeout, ca_path, unredirected_headers,
+                  decompress, ciphers, use_netrc)
 
     elapsed = (datetime.datetime.utcnow() - start).seconds
 
